#---------------------------------------------------------------------
#
# licensed under the terms of GNU GPL 2
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
#---------------------------------------------------------------------

import os,sys
sys.path.append("/usr/share/qgis/python")
from PyQt4.QtCore import *
from PyQt4.QtGui import *
from qgis.core import *

from ui_control import ui_Control
import resources
from math import *
from getcoordtool import *

class tarsius (object):
    """
    Base class for the tarsius plugin
    - Provides a means to draw a feature by specifying the angle and distance beetween points.
    - Supports angles in either the conventional 0.0 - 360.0 clockwise from North
        or the surveyor's 'Easting' system with bearings plus or minus 90 deg. from North or South
    - Supports magnetic declination as degrees plus or minus for East or West respectively
    - supports inputs in feet or the current CRS units
    Based on qgsAzimuth plugin.
    """

    # just a test to see if mods are taking

    def __init__(self, iface):
        self.iface = iface
        self.canvas = iface.mapCanvas()
        self.fPath = QString()  # set default working directory, updated from config file & by Import/Export

    def initGui(self):
        # create action that will start plugin configuration
        self.action = QAction(QIcon(":qgsazimuth.png"), "Tarsius", self.iface.mainWindow())
        self.action.setWhatsThis("Tarsius plugin")

        # chytneme F7
        #self.keyAction = QAction("Tarsius Plugin", self.iface.mainWindow())
        self.iface.registerMainWindowAction(self.action, "F7") # action1 is triggered by the F7 key
	self.iface.addPluginToMenu("&Telemetry", self.action)
        QObject.connect(self.action, SIGNAL("triggered()"),self.run)

        # add toolbar button and menu item
        self.iface.addPluginToMenu("&Telemetry", self.action)
        self.pluginGui = ui_Control(self.iface.mainWindow())

        self.tool = GetCoordTool(self.canvas)

    def unload(self):
        # remove the plugin menu item and icon
        self.iface.removePluginMenu("&Telemetry",self.action)
        self.iface.removeToolBarIcon(self.action)
        self.saveConf()
	self.fBaseName = ''

    def run(self):
        # create and show a configuration dialog or something similar
        flags = Qt.WindowTitleHint | Qt.WindowSystemMenuHint | Qt.WindowMaximizeButtonHint  # QgisGui.ModalDialogFlags
        self.pluginGui = ui_Control(self.iface.mainWindow())

        #misc init
        self.loadConf() # get config data
        self.clearList()
        self.filter_fix = None
        self.filter_i = None

        #INSERT EVERY SIGNAL CONECTION HERE!
        QObject.connect(self.pluginGui.pushButton_segListRowDel,SIGNAL("clicked()"),self.delRow)
        QObject.connect(self.pluginGui.pushButton_segListLoad,SIGNAL("clicked()"),self.loadList)
        QObject.connect(self.pluginGui.pushButton_segListClear,SIGNAL("clicked()"),self.clearList)
        QObject.connect(self.pluginGui.pushButton_clearLines,SIGNAL("clicked()"),self.clearLines)
        QObject.connect(self.pluginGui.pushButton_objectDraw,SIGNAL("clicked()"),self.drawSingle)
        QObject.connect(self.pluginGui.pushButton_objectDrawFix,SIGNAL("clicked()"),self.drawFix)
        QObject.connect(self.pluginGui.pushButton_objectDrawAll,SIGNAL("clicked()"),self.drawAll)
        QObject.connect(self.pluginGui.pushButton_addFixLayer,SIGNAL("clicked()"),self.addFixLayer)
        QObject.connect(self.pluginGui.pushButton_addFix,SIGNAL("clicked()"),self.addFix)
        QObject.connect(self.pluginGui.pushButton_save,SIGNAL("clicked()"),self.saveFixes)
        QObject.connect(self.pluginGui.comboBox_layers,SIGNAL("activated(QString)"),self.layerSelected)

        QObject.connect(self.pluginGui, SIGNAL("finished(int)"), self.close)
	
        # Initialize points combo box
	self.points = None
        self.layermap=QgsMapLayerRegistry.instance().mapLayers()
        activeName = ""
        self.pluginGui.comboBox_layers.addItem("------")
        for (name,layer) in sorted(self.layermap.iteritems()):
            if layer.type() != layer.VectorLayer or layer.dataProvider().geometryType() != 1:
                # allow only POINTS layers
                continue
            self.pluginGui.comboBox_layers.addItem(layer.name())
            #if (layer == self.iface.activeLayer()):
#            if (name == self.lines_layer):
#                self.pluginGui.lineEdit_crs.setText((layer.srs()).description())
#                #self.say('found active layer='+name)
#                activeName = name
#                # set combo box to current active layer
#                lyrNdx = self.pluginGui.comboBox_layers.findText(activeName)
#                self.pluginGui.comboBox_layers.setCurrentIndex(lyrNdx)

	# addFix button is only enabled if a points layer is selected
	if not self.points:
        	self.pluginGui.pushButton_addFix.setEnabled(False)

        # Initialize waipoints layer combo box
        self.layermap=QgsMapLayerRegistry.instance().mapLayers()
        activeName = ""
        for (name,layer) in sorted(self.layermap.iteritems()):
            if layer.type() != layer.VectorLayer or layer.dataProvider().geometryType() != 1:
                # allow only POINTS layers
                continue
            self.pluginGui.comboBox_waypoints.addItem(layer.name())
            if (layer.name() == self.waypointsLayerName):
                activeName = layer.name()
                # set combo box to current active layer
                lyrNdx = self.pluginGui.comboBox_waypoints.findText(activeName)
                self.pluginGui.comboBox_waypoints.setCurrentIndex(lyrNdx)


        self.pluginGui.table_segmentList.setCurrentCell(0,0)
        self.pluginGui.show()

	# create a vecor layer for drawing lines
	self.lines = None
	for (name, layer) in self.layermap.iteritems():
		if layer.name() == 'temporary_lines':
			self.lines = layer
	if not self.lines:
		self.lines = QgsVectorLayer("LineString", "temporary_lines", "memory")
       		renderer=self.canvas.mapRenderer()
		self.lines.setCrs(renderer.destinationSrs())
		pr = self.lines.dataProvider()
		self.lines.startEditing()

		# add label attribute
		pr.addAttributes( [ QgsField("label", QVariant.String) ] )
		QgsMapLayerRegistry.instance().addMapLayer(self.lines)
		self.lines.commitChanges()

		# enable label
		label = self.lines.label()
		label.setLabelField(0,0)

	self.lines.startEditing()
	
    def close(self):
        self.saveConf()

    #Now these are the SLOTS
    def nextvertex(self,v,d,az,zen=90):
        az=radians(az)
        zen=radians(zen)
        d1=d*sin(zen)
        x=v[0]+d1*sin(az)
        y=v[1]+d1*cos(az)
        z=v[2]+d*cos(zen)
        return [x,y,z]

    def drawSingle(self):
        cur_row = self.pluginGui.table_segmentList.currentRow()
        if cur_row < 0:
            self.say("No record was selected")
            return 0
        self.add_one_geometry(cur_row)

    def drawFix(self):
        cur_row = self.pluginGui.table_segmentList.currentRow()
        if cur_row < 0:
            self.say("No record was selected")
            return 0
        for i in range(self.pluginGui.table_segmentList.rowCount()):
	    if self.get_fix(i) == self.get_fix(cur_row):
		self.add_one_geometry(i)

    def drawAll(self):
        for i in range(self.pluginGui.table_segmentList.rowCount()):
            self.add_one_geometry(i)

    def add_one_geometry(self, i):
        self.iface.setActiveLayer(self.lines)
        stat=str(self.pluginGui.table_segmentList.item(i,2).text())
	if not stat:
            self.say('Station missing on line %d.' % (i+1))
	    return -1
	    
        az=str(self.pluginGui.table_segmentList.item(i,3).text())
	if not az:
            self.say('Azimuth missing on line %d.' % (i+1))
	    return -1

        dis=130 #XXX
        zen="90"


        waypoints = self.getVectorLayerByName(self.pluginGui.comboBox_waypoints.currentText())
        feature = self.findFeature(stat)
        if not feature:
           return -1

        v0 = feature.geometry().asPoint()
        repr = self.reproject([v0], waypoints)[0]
	repr_coord = [repr.x(), repr.y(), 90]

        az=float(self.dmsToDd(az))
        zen=float(self.dmsToDd(zen))

        #correct for angles outside of 0.0-360.0
        while (az > 360.0):
            az = az - 360.0
        while (az < 0.0):
            az = az + 360.0

        new_point_coord = self.nextvertex(repr_coord,dis,az,zen)
	new_point = QgsPoint(new_point_coord[0], new_point_coord[1])

        featurelist=[]
        pointlist=[]
        stat=self.get_stat(i)
        time=str(self.pluginGui.table_segmentList.item(i,1).text())
        #writing new feature
        pointlist=[repr,new_point]
        geom=QgsGeometry.fromPolyline(pointlist)
        feature=QgsFeature()
        feature.setGeometry(geom)
        b = QVariant(time)
        feature.changeAttribute(0, b)
        featurelist.append(feature)

        #commit
        self.lines.addFeatures(featurelist)
	self.lines.select(self.lines.extent(), True)
        self.iface.mapCanvas().zoomToSelected()

    def dmsToDd(self,dms):
        "It's not fast, but it's a safe way of dealing with DMS"
        dms=dms.replace(" ", "")
        for c in dms:
            if ((not c.isdigit()) and (c != '.') and (c != '-')):
                dms=dms.replace(c,';')
        while (dms.find(";;")>=0):
            dms=dms.replace(";;",';')
        if dms[0]==';':
            dms=dms[1:]
        dms=dms.split(";")
        dd=0
        #dd=str(float(dms[0])+float(dms[1])/60+float(dms[2])/3600)
        for i, f in enumerate(dms):
            if f!="":
                dd+=float(f)/pow(60, i)
        return dd

    def clearLines(self):
	self.lines.select(self.lines.extent(), True)
	self.lines.deleteSelectedFeatures()

    def saveFixes(self):
	self.points.commitChanges()

    def clearList(self):
        self.pluginGui.table_segmentList.clearContents()
        self.pluginGui.table_segmentList.setSortingEnabled(False)
        self.pluginGui.table_segmentList.setRowCount(0)
        self.pluginGui.table_segmentList.setCurrentCell(0, 0) # substitute for missing setCurrentRow()
        #retranslateUi


    def addFixLayer(self):

	# Generate a save file dialog with a dropdown box for choosing encoding style
	def saveDialog(parent):
		settings = QSettings()
		dirName = settings.value("/UI/lastShapefileDir").toString()
		filtering = QString("Shapefiles (*.shp)")
		encode = settings.value("/UI/encoding").toString()
		fileDialog = QgsEncodingFileDialog(parent, "Save output shapefile", dirName, filtering, encode)
		fileDialog.setDefaultSuffix(QString("shp"))
		fileDialog.setFileMode(QFileDialog.AnyFile)
		fileDialog.setAcceptMode(QFileDialog.AcceptSave)
		fileDialog.setConfirmOverwrite(True)
		if not fileDialog.exec_() == QDialog.Accepted:
				return None, None
		files = fileDialog.selectedFiles()
		settings.setValue("/UI/lastShapefileDir", QVariant(QFileInfo(unicode(files.first())).absolutePath()))
		return (unicode(files.first()), unicode(fileDialog.encoding()))

	(self.shapefileName, self.encoding) = saveDialog(None)
	if self.shapefileName is None or self.encoding is None:
		return # XXX

	fields = {
		0: QgsField("animal",  	QVariant.String),
		1: QgsField("datetime", QVariant.String),
		3: QgsField("quality",  QVariant.Int),
	}
	geometryType = QGis.WKBPoint # POINTS
       	srs = self.canvas.mapRenderer().destinationSrs()
	fixLayerName = os.path.basename(self.shapefileName)
	fixLayerName = fixLayerName[:-4] # strip '.shp'

	check = QFile(self.shapefileName)
	if check.exists():
		if not QgsVectorFileWriter.deleteShapeFile(self.shapefileName):
			return
	writer = QgsVectorFileWriter(self.shapefileName, self.encoding, fields, geometryType, srs)
	if writer.hasError() != QgsVectorFileWriter.NoError:
		self.tell("Error when creating shapefile: ", writer.hasError())
	del writer

	layer = QgsVectorLayer(self.shapefileName, fixLayerName, "ogr")
	layer.startEditing()
	QgsMapLayerRegistry.instance().addMapLayer(layer)

	self.pluginGui.comboBox_layers.addItem(layer.name())

	# set combo box to current active layer
	lyrNdx = self.pluginGui.comboBox_layers.findText(layer.name())
	self.pluginGui.comboBox_layers.setCurrentIndex(lyrNdx)
	# setCurrentIndex does not emit selected() signal
	self.layerSelected(layer.name())

    def addRow(self,  az=0,  dist=0,  zen = 90, day='', time='', stat='', fix='', note=''):
        #add the vertext to the end of the table
        i=self.pluginGui.table_segmentList.rowCount()
        self.pluginGui.table_segmentList.insertRow(i)
        self.pluginGui.table_segmentList.setItem(i, 0, QTableWidgetItem(str(day)))
        self.pluginGui.table_segmentList.setItem(i, 1, QTableWidgetItem(str(time)))
        self.pluginGui.table_segmentList.setItem(i, 2, QTableWidgetItem(str(stat)))
        self.pluginGui.table_segmentList.setItem(i, 3, QTableWidgetItem(str(az)))
        self.pluginGui.table_segmentList.setItem(i, 4, QTableWidgetItem(str(fix)))
        self.pluginGui.table_segmentList.setItem(i, 5, QTableWidgetItem(str(note)))

    def insertRow(self):
        az = self.pluginGui.lineEdit_nextAzimuth.text()
        dist = self.pluginGui.lineEdit_nextDistance.text()
        zen = self.pluginGui.lineEdit_nextVertical.text()

        #insert the vertext into the table at the current position
        i=self.pluginGui.table_segmentList.currentRow()
        self.pluginGui.table_segmentList.insertRow(i)
        self.pluginGui.table_segmentList.setItem(i, 0, QTableWidgetItem(str(az).upper()))
        self.pluginGui.table_segmentList.setItem(i, 1, QTableWidgetItem(str(dist)))
        self.pluginGui.table_segmentList.setItem(i, 2, QTableWidgetItem(str(zen)))

    def delRow(self):
        self.pluginGui.table_segmentList.removeRow(self.pluginGui.table_segmentList.currentRow())

    def moveup(self):
        pass

    def movedown(self):
        pass

    def addFix(self):
        self.iface.setActiveLayer(self.points)
        QObject.connect(self.tool, SIGNAL("finished(PyQt_PyObject)"), self.getpoint)
        self.saveTool = self.canvas.mapTool()
        self.canvas.setMapTool(self.tool)

    def getpoint(self,pt):
        self.point = pt

        feature = QgsFeature()
        geometry = QgsGeometry()
        feature.setGeometry(geometry.fromPoint(self.point))
        self.canvas.setMapTool(self.saveTool)

        # we expect the date fromat  dd.mm.yyyy
        # and time format HH:MM (24 hour)
        # e.g. 17.4.2010 03:25
        cur_row = self.pluginGui.table_segmentList.currentRow()
        day=str(self.pluginGui.table_segmentList.item(cur_row,0).text())
        time=str(self.pluginGui.table_segmentList.item(cur_row,1).text())
        from datetime import datetime
        dt = datetime.strptime(day + ' ' + time, '%d.%m.%Y %H:%M')
        # convert to format required by Time Manager plugin
	dt_str = dt.strftime('%Y-%m-%d %H:%M')
        feature.changeAttribute(1, dt_str)

        # try to parse animal ID from the filename
	# e.g. F36_0608.csv means animal 'F36'
	if self.fBaseName:
		animal_id = self.fBaseName.split('_')[0]
        	feature.changeAttribute(0, animal_id)

        ok = self.iface.openFeatureForm(self.points, feature)
	if ok:
		self.points.beginEditCommand("Add fix")
	        self.points.addFeature(feature)
		self.points.endEditCommand()
        self.canvas.refresh()
        QObject.disconnect(self.tool, SIGNAL("finished(PyQt_PyObject)"), self.getpoint)

    def layerSelected(self, name):
	self.points = self.getVectorLayerByName(name)
	self.points.startEditing()
	self.iface.setActiveLayer(self.points)
        self.pluginGui.pushButton_addFix.setEnabled(True)


    def reproject(self, vlist,  vectorlayer):
	srs = QgsCoordinateReferenceSystem(4326)
	vectorlayer.setCrs(srs)
        renderer=self.canvas.mapRenderer()
        for i, point in enumerate(vlist):
            vlist[i]= renderer.layerToMapCoordinates(vectorlayer, point)
        return vlist

    def setAngle(self, s):
        if (s=='azimuth'):
            self.pluginGui.radioButton_azimuthAngle.setChecked(True)
        elif (s=='bearing'):
            self.pluginGui.radioButton_bearingAngle.setChecked(True)
        elif (s=='polar'):
            self.pluginGui.radioButton_polorCoordAngle.setChecked(True)
        else:
            self.say('invalid angle type: '+s)

    def setHeading(self,  s):
        if (s=='coordinate_system'):
            self.pluginGui.radioButton_defaultNorth.setChecked(True)
        elif (s=='magnetic'):
            self.pluginGui.radioButton_magNorth.setChecked(True)
        else:
            self.say('invalid heading type: '+s)

    def say(self, txt):
        # present a message box on screen
        warn=QgsMessageViewer()
        warn.setMessageAsPlainText(txt)
        warn.showMessage()

    def tell(self, txt):
        # write to bottom of Note area at top of screen
        #self.notes.appendPlainText(str(txt))
	self.say(txt) # note area was removed

    # ---------------------------------------------------------------------------------------------------------------------------------
    #               File handling
    # This section deals with saving the user data to disk, and loading it
    #
    # format: TODO
    # ---------------------------------------------------------------------------------------------------------------------------------
    def loadList(self):
        self.fileName=QFileDialog.getOpenFileName(None,"Load data separated by ';'",self.fPath, "Telemetry data (*.csv)", None)
        if not os.path.exists(self.fileName):
            return 0
        # update selected file's folder
        fInfo = QFileInfo(self.fileName)
        self.fPath = fInfo.absolutePath()
	self.fBaseName = fInfo.baseName()
        self.saveConf()

        # get saved data
        f=open(self.fileName)
        lines=f.readlines()
        f.close()
        self.clearList()
	errcount = 0
	linenum = 0
        for line in lines:
	    linenum += 1
            #remove trailing 'new lines', etc and break into parts
            parts = ((line.strip()).lower()).split("=")
            if (len(parts)>1):
                #self.say("line="+line+'\nparts[0]='+parts[0]+'\nparts[1]='+parts[1])
#                if (parts[0].lower()=='angle'):
#                    self.setAngle(parts[1].lower())
#                elif (parts[0].lower()=='heading'):
#                    self.setHeading(parts[1].lower())
#                elif (parts[0].lower()=='declination'):
#                    self.setDeclination(parts[1].lower())
#                elif (parts[0].lower()=='dist_units'):
#                    self.setDistanceUnits(parts[1].lower())
#                elif (parts[0].lower()=='startat'):
#                    self.setStartAt(parts[1].lower())
#                elif (parts[0].lower()=='survey'):
#                    self.setSurvey(parts[1].lower())
                pass
            else:
                coords=(line.strip()).split(";")
		# detect the heading
                if (coords[7].lower()=='fix'):
                    pass
                else:
		    day = coords[2]
		    time = coords[3]
		    stat = coords[4]
		    az = coords[6]
		    fix = coords[7]
		    try:
                        note = coords[8]
                    except IndexError:
                        note = '' 
		    if not day and not time and not stat and not az and not fix and not note:
			    # skip empty line
			    continue
                    self.addRow(day=coords[2], time=coords[3], stat=stat, az=az, fix=fix, note=note)
#	if errcount:
#            self.say('There were %d errors in the file. See message pane for details' % errcount)


    #------------------------
    def loadConf(self):
        settings=QSettings()
        size = settings.value('/Plugin-tarsius/size', QVariant(QSize(800, 600))).toSize()
        self.pluginGui.resize(size)
        position = settings.value('/Plugin-tarsius/position', QVariant(QPoint(0, 0))).toPoint()
        self.pluginGui.move(position)
        #settings.restoreGeometry(settings.value("Geometry").toByteArray())
        self.fPath = settings.value('/Plugin-tarsius/inp_exp_dir').toString()
        self.fileName = self.fPath
        self.lines_layer = settings.value('/Plugin-tarsius/lines_layer').toString()
	self.waypointsLayerName = settings.value('/Plugin-tarsius/waypoints_layer').toString()

    def saveConf(self):
        settings=QSettings()
        #settings.setValue("Geometry", QVariant(self.saveGeometry()))
        settings.setValue('/Plugin-tarsius/size',  QVariant(self.pluginGui.size()))
        settings.setValue('/Plugin-tarsius/position',  QVariant(self.pluginGui.pos()))
        settings.setValue('/Plugin-tarsius/inp_exp_dir', QVariant(self.fPath))
        settings.setValue('/Plugin-tarsius/lines_layer', QVariant(self.pluginGui.comboBox_layers.currentText()))
        settings.setValue('/Plugin-tarsius/waypoints_layer', QVariant(self.pluginGui.comboBox_waypoints.currentText()))

    def sortedDict(self, adict):
        keys = adict.keys()
        keys.sort()
        return map(adict.get, keys)

    def findFeature(self, name):
        l = self.getVectorLayerByName(self.pluginGui.comboBox_waypoints.currentText())
        #l = self.iface.activeLayer()
        provider = l.dataProvider()
        allAttrs = provider.attributeIndexes()
        provider.select(allAttrs)
        feat = QgsFeature()
        while provider.nextFeature(feat):
            attrs = feat.attributeMap()
            if str(attrs[0].toString()).lower() == name.lower():
                    return feat
        self.say("Waypoint %s was not found" % name)
        return 0

    def getVectorLayerByName(self, name):
    	for i in range(self.canvas.layerCount()):
		mapLayer = self.canvas.layer(i)
		if mapLayer.name() == name:
			return  mapLayer
	'''
	nLayers = self.canvas.layerCount()
        for l in range(nLayers):
            layer = self.canvas.layer(l)
            if layer.name() == unicode(myName):
                vlayer = QgsVectorLayer(unicode(layer.source()),  unicode(myName),  unicode(layer.dataProvider().name()))
                if vlayer.isValid():
                    return vlayer
                else:
                    self.say("Vector layer is not valid")
	'''

    def get_stat(self, i):
        return str(self.pluginGui.table_segmentList.item(i,2).text())
    def get_fix(self, i):
        return str(self.pluginGui.table_segmentList.item(i,4).text())
        
